
from libcloud.compute.types import NodeState

from plumbery.engine import PlumberyEngine

# report on the hard work
aliensCount = 0
nodesCount = 0

# load the full description of fittings
plumbery = PlumberyEngine('fittings.yaml')

# look at each facility
for facility in plumbery.facilities:
	facility.focus()

	# enumerate domains described in blueprints
	citizenDomainNames = set()
	for blueprint in facility.fittings.blueprints:
		attributes = blueprint.values()[0]
		citizenDomainNames.add(attributes['domain']['name'])

	# turn these names to actual unique ids
	citizenDomainIds = []
	for domain in facility.region.ex_list_network_domains(location=facility.location):
		if domain.name in citizenDomainNames:
			citizenDomainIds.append(domain.id)

	# enumerate nodes described in blueprints
	citizenNodeNames = set()
	for blueprint in facility.fittings.blueprints:
		for item in blueprint.values()[0]['nodes']:
			citizenNodeNames.add(item.keys()[0])

	# now we can look at each deployed node
	for node in facility.region.list_nodes():

		# skip nodes from other facilities
		if node.extra['datacenterId'] != facility.location.id:
			continue

		# work begin
		nodesCount += 1

		# node is part of our fittings -- good boy
		if node.name in citizenNodeNames:
			continue

		# node is part of a citizen network domain -- good boy too
		if node.extra['networkDomainId'] in citizenDomainIds:
			continue

		# we found an alien
		aliensCount += 1

		# offer to shutdown running alien nodes
		if node.state is NodeState.RUNNING:
			print("'{}' has been deployed is not part of our fittings".format(node.name))
			feedback = raw_input('Should this node been stopped? [y/N/q] ').lower()

			# we are asked to shutdown this node
			if len(feedback) > 0 and feedback[0] == 'y':

				if plumbery.safeMode:
					print("Would have stopped node '{}' if not in safe mode".format(node.name))

				else:
					print("Stopping node '{}'".format(node.name))

					# we may have to wait for busy resources
					while True:

						try:
							facility.region.ex_shutdown_graceful(plumbery.region.ex_get_node_by_id(node.id))
							print "- in progress"

						except Exception as feedback:

							# resource is busy, wait a bit and retry
							if 'RESOURCE_BUSY' in str(feedback):
								facility.wait_and_tick()
								continue

							# other exception
							else:
								print("Error: unable to stop node '{}'!".format(node.name))
								print feedback

						# quit the loop
						break


			# quit this loop immediately
			elif len(feedback) > 0 and feedback[0] == 'q':
				break

# report on the hard work
print("{} nodes have been processed in total, including {} alien nodes".format(nodesCount, aliensCount))







#!/usr/bin/env python

"""
Tests for `facility` module.
"""

import os
import unittest

from plumbery.engine import PlumberyBlueprints
from plumbery.facility import PlumberyFacility


class FakeImage:

    name = 'RedHat 6 64-bit 4 CPU'

class FakeLocation:

    id = 'EU7'
    name = 'data centre in Amsterdam'
    country = 'Netherlands'


class FakePlumbery:

    userName = 'n'
    userPassword = 'p'
    safeMode = False
    sharedSecret = 's'

    def driver(self, name, password, region):
        return FakeRegion()


class FakeNode:

    name = 'stackstorm'
    extra = {'datacenterId': 'EU7'}


class FakeRegion:

    def create_node(self, name, image, auth, ex_network_domain, ex_vlan, ex_is_started, ex_description):
        return True

    def ex_create_network_domain(self, location, name, service_plan, description):
        return True

    def ex_create_vlan(self, network_domain, name, private_ipv4_base_address, description):
        return True

    def destroy_node(self, node):
        return True

    def ex_get_location_by_id(self, location):
        return FakeLocation()

    def ex_list_network_domains(self, location):
        return []

    def ex_list_vlans(self, location, network_domain):
        return []

    def ex_start_node(self, node):
        return True

    def ex_shutdown_graceful(self, node):
        return True

    def list_images(self, location):
        return [FakeImage()]

    def list_nodes(self):
        return [FakeNode()]


fakeFittings = {
    'regionId': 'dd-eu',
    'locationId': 'EU7',
    'blueprints': [{
            'fake': {
                    'domain': {
                            'name': 'VDC1',
                            'service': 'ADVANCED',
                            'description': 'fake'},
                    'ethernet': {
                            'name': 'vlan1',
                            'subnet': '10.0.10.0',
                            'description': 'fake'},
                     'nodes': [{
                            'stackstorm': {
                                    'description': 'fake',
                                    'appliance': 'RedHat 6 64-bit 4 CPU'
                                    }
                            }]
                    }
            }]
    }


class TestPlumberyFacility(unittest.TestCase):

    def setUp(self):
        plumbery = FakePlumbery()
        self.fittings = PlumberyBlueprints(**fakeFittings)
        self.facility = PlumberyFacility(plumbery=plumbery, fittings=self.fittings)

    def tearDown(self):
        self.facility = None

    def test_000(self):
        self.facility.build_all_blueprints()

    def test_001(self):
        self.facility.build_blueprint('fake')

    def test_010(self):
        self.facility.destroy_all_nodes()

    def test_011(self):
        self.facility.destroy_nodes('fake')

    def test_020(self):
        self.facility.focus()

    def test_030(self):
        self.facility.get_blueprint('fake')

    def test_031(self):
        self.facility.get_node('stackstorm')

    def test_040(self):
        self.facility.start_all_nodes()

    def test_041(self):
        self.facility.start_nodes('fake')

    def test_050(self):
        self.facility.stop_all_nodes()

    def test_051(self):
        self.facility.stop_nodes('fake')

    def test_099(self):
        self.facility.wait_and_tick(0)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python

"""
Tests for `polisher` module.
"""

import unittest

from libcloud.compute.types import NodeState

from plumbery.polisher import PlumberyPolisher


class TestPlumberyPolisher(unittest.TestCase):

    def setUp(self):
        self.polisher = PlumberyPolisher.from_shelf('spit')

    def tearDown(self):
        self.polisher = None

    def testPolish(self):

        class FakeNode(object):
            name = 'fake'
            state = NodeState.RUNNING
            private_ips = ['10.100.100.100']

        self.polisher.shine_node(FakeNode())


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# standard libraries
import sys
import time

# Apache Libcloud - https://libcloud.readthedocs.org/en/latest
from libcloud.compute.base import NodeAuthPassword
from libcloud.compute.types import NodeState

# other code related to plumbery
from domain import PlumberyDomain
from polisher import PlumberyPolisher


__all__ = ['PlumberyFacility']


class PlumberyFacility:
    """Plumbing at one facility

    Args:

        plumbery (PlumberyEngine): the automate that is coordinating
            plumbing activities at multiple facilities

        fittings (PlumberyBlueprints): the plan for the fittings

    Example::

        from plumbery.facility import PlumberyFacility
        facility = PlumberyFacility(plumbery=plumbery, fittings=fittings)
        facility.build_all_blueprints()

    In this example a facility is ruled by the given plumber, and the plan
    of all the fittings is provided as well.

    Attributes:

        plumbery: global parameters and functions

        fittings: the plan is available when needed

    """

    # the description of the fittings
    fittings = None

    # the list of available images to create nodes
    images = None

    # the target physical data center
    location = None

    # the handle to global parameters and functions
    plumbery = None

    # the handle to the Apache Libcloud driver
    region = None

    def __init__(self, plumbery=None, fittings=None):
        """Put this facility in context"""

        # handle to global parameters and functions
        self.plumbery = plumbery

        # parameters for this location
        self.fittings = fittings

        # configure the API endpoint - regional parameter is related to federated structure of cloud services at Dimension Data
        self.region = plumbery.driver(plumbery.userName, plumbery.userPassword, region=fittings.regionId)

        # focus at one specific location - and attempt to use the API over the network
        try:
            self.location = self.region.ex_get_location_by_id(fittings.locationId)
        except Exception as feedback:
            print("Error: unable to communicate with API endpoint - have you checked http_proxy environment variable?")
            print(str(feedback))
            exit(-1)

        # fetch the list of available images only once from the API
        self.images = self.region.list_images(location=self.location)

    def build_all_blueprints(self):
        """Build all blueprints"""

        for blueprint in self.fittings.blueprints:
            print("Building blueprint '{}'".format(blueprint.keys()[0]))
            self.build_blueprint(blueprint.keys()[0])

    def build_blueprint(self, name):
        """Build a named blueprint"""

        # get the blueprint
        blueprint = self.get_blueprint(name)
        if not blueprint:
            return

        # create the network domain if it does not exist
        domain = PlumberyDomain(self)
        domain.build(blueprint)

        # create nodes that do not exist
        self._build_nodes(blueprint=blueprint, domain=domain)

    def _build_nodes(self, blueprint, domain):
        """Create nodes if they do not exist"""

        # ensure that we have some nodes described here
        if 'nodes' not in blueprint:
            print("Error: no nodes have been defined for the blueprint '{}'!".format(blueprint['target']))
            exit(-1)

        # respect the order of nodes defined in the fittings description
        for item in blueprint['nodes']:

            # node has several explicit attributes
            if type(item) is dict:
                nodeName = item.keys()[0]
                nodeAttributes = item.values()[0]

            # node has only a name
            else:
                nodeName = item
                nodeAttributes = None

            # node may already exist
            if self.get_node(nodeName):
                print("Node '{}' already exists".format(nodeName))

            # create a new node
            else:

                # the description attribute is a smart way to tag resources
                description = '#plumbery'
                if type(nodeAttributes) is dict and 'description' in nodeAttributes:
                    description = nodeAttributes['description'] + ' #plumbery'

                # define which appliance to use
                if type(nodeAttributes) is dict and 'appliance' in nodeAttributes:
                    imageName = nodeAttributes['appliance']
                else:
                    imageName = 'Ubuntu'

                # find suitable image to use
                image = None
                for image in self.images:
                    if imageName in image.name:
                        break

                # Houston, we've got a problem
                if image is None or imageName not in image.name:
                    print("Error: unable to find image for '{}'!".format(imageName))
                    exit(-1)

                # safe mode
                if self.plumbery.safeMode:
                    print("Would have created node '{}' if not in safe mode".format(nodeName))

                # actual node creation
                else:
                    print("Creating node '{}'".format(nodeName))

                    # we may have to wait for busy resources
                    while True:

                        try:
                            self.region.create_node(
                                name=nodeName,
                                image=image,
                                auth=NodeAuthPassword(self.plumbery.sharedSecret),
                                ex_network_domain=domain.domain,
                                ex_vlan=domain.network,
                                ex_is_started=False,
                                ex_description=description)
                            print("- in progress")

                        except Exception as feedback:

                            # resource is busy, wait a bit and retry
                            if 'RESOURCE_BUSY' in str(feedback):
                                self.wait_and_tick()
                                continue

                            # fatal error
                            else:
                                print("Error: unable to create node '{}'!".format(nodeName))
                                print(str(feedback))
                                exit(-1)

                        # quit the loop
                        break

    def destroy_all_nodes(self):
        """Destroy all nodes"""

        # destroy in reverse order
        for blueprint in self.fittings.blueprints:
            self.destroy_nodes(blueprint.keys()[0])

    def destroy_nodes(self, name):
        """Destroy nodes"""

        # get the blueprint
        blueprint = self.get_blueprint(name)
        if not blueprint:
            return

        # ensure that some nodes have been described
        if 'nodes' not in blueprint:
            return

        # destroy in reverse order
        for item in reversed(blueprint['nodes']):

            # find the name of the node to be destroyed
            if type(item) is dict:
                nodeName = item.keys()[0]
            else:
                nodeName = str(item)

            # enumerate existing nodes
            node = self.get_node(nodeName)

            # destroy an existing node
            if node is not None:

                # safe mode
                if self.plumbery.safeMode:
                    print("Would have destroyed node '{}' if not in safe mode".format(nodeName))

                # actual node destruction
                else:
                    print("Destroying node '{}'".format(nodeName))

                    # we may have to wait for busy resources
                    while True:

                        try:
                            self.region.destroy_node(node)
                            print("- in progress")

                        except Exception as feedback:

                            # resource is busy, wait a bit and retry
                            if 'RESOURCE_BUSY' in str(feedback):
                                self.wait_and_tick()
                                continue

                            # node is up and running, would have to stop it first
                            elif 'SERVER_STARTED' in str(feedback):
                                print("- skipped - node is up and running")

                            # fatal error
                            else:
                                print("Error: unable to destroy node '{}'!".format(nodeName))
                                print(str(feedback))
                                exit(-1)

                        # quit the loop
                        break

    def focus(self):
        """Where are we plumbing?"""

        print("Plumbing at '{}' {} ({})".format(self.location.id, self.location.name, self.location.country))

    def get_blueprint(self, name):
        """Get a blueprint by name"""

        for blueprint in self.fittings.blueprints:
            if name in blueprint.keys()[0]:
                blueprint = blueprint[name]
                blueprint['target'] = name
                return blueprint

        return None

    def get_node(self, name):
        """Get a node by name"""

        # enumerate existing nodes
        node = None
        for node in self.region.list_nodes():

            # skip nodes from other locations
            if node.extra['datacenterId'] != self.location.id:
                continue

#           # skip nodes from other network domains
#           if node.extra['networkDomainId'] != self.domain.id:
#               continue

            # found an existing node with this name
            if node.name == name:
                return node

        # not found
        return None

    def polish_node(self, node, polisher):
        """Wait for a node to be started and polish it"""

        # we have to wait until node is running
        while node is not None:
            if node.state == NodeState.RUNNING:
                break

            # give it some time to start
            PlumberyFacility.wait_and_tick()
            node = self.get_node(node.name)
            continue

        # polish
        if node:
            polisher.shine_node(node)

    def start_all_nodes(self):
        """Start all nodes"""

        for blueprint in self.fittings.blueprints:
            self.start_nodes(blueprint.keys()[0])

    def _start_node(self, name, attributes=None):
        """Start a node"""

        # get fresh state of the node
        node = self.get_node(name)

        # not found
        if node is None:
            return None

        # define which polish to use
        polisher = None
        if type(attributes) is dict and 'polish' in attributes:
            polish = attributes['polish']
            polisher = PlumberyPolisher.from_shelf(polish)

        # safe mode
        if self.plumbery.safeMode:
            print("Would have started node '{}' if not in safe mode".format(name))

        # actual node start
        else:
            print("Starting node '{}'".format(name))

            # we may have to wait for busy resources
            while True:

                try:
                    self.region.ex_start_node(node)
                    print("- in progress")

                    # if there is a need to polish the appliance, we may have to wait a bit more
                    if polisher:
                        self.polish_node(node, polisher)

                except Exception as feedback:

                    # resource is busy, wait a bit and retry
                    if 'RESOURCE_BUSY' in str(feedback):
                        self.wait_and_tick()
                        continue

                    # node is up and running, nothing to do
                    elif 'SERVER_STARTED' in str(feedback):
                        print("- skipped - node is up and running")

                    # fatal error
                    else:
                        print("Error: unable to start node '{}'!".format(name))
                        print(str(feedback))
                        exit(-1)

                # quit the loop
                break

        # provide more details
        return node

    def start_nodes(self, name):
        """Start nodes"""

        # get the blueprint
        blueprint = self.get_blueprint(name)
        if not blueprint:
            return

        # ensure that some nodes have been described
        if 'nodes' not in blueprint:
            return

        # start nodes
        for item in blueprint['nodes']:

            # node has several explicit attributes
            if type(item) is dict:
                nodeName = item.keys()[0]
                nodeAttributes = item.values()[0]

            # node has only a name
            else:
                nodeName = item
                nodeAttributes = None

            # one node at at time
            self._start_node(nodeName, nodeAttributes)

    def stop_all_nodes(self):
        """Stop all nodes"""

        for blueprint in self.fittings.blueprints:
            self.stop_nodes(blueprint.keys()[0])

    def stop_nodes(self, name):
        """Stop nodes"""

        # get the blueprint
        blueprint = self.get_blueprint(name)
        if not blueprint:
            return

        # ensure that some nodes have been described
        if 'nodes' not in blueprint:
            return

        # stop nodes
        for item in blueprint['nodes']:

            # find the name of the node to be destroyed
            if type(item) is dict:
                nodeName = item.keys()[0]
            else:
                nodeName = str(item)

            # enumerate existing nodes
            node = self.get_node(nodeName)

            # stop an existing node
            if node is not None:

                # safe mode
                if self.plumbery.safeMode:
                    print("Would have stopped node '{}' if not in safe mode".format(nodeName))

                # actual node stop
                else:
                    print("Stopping node '{}'".format(nodeName))

                    # we may have to wait for busy resources
                    while True:

                        try:
                            self.region.ex_shutdown_graceful(node)
                            print("- in progress")

                        except Exception as feedback:

                            # resource is busy, wait a bit and retry
                            if 'RESOURCE_BUSY' in str(feedback):
                                self.wait_and_tick()
                                continue

                            # transient error, wait a bit and retry
                            elif 'UNEXPECTED_ERROR' in str(feedback):
                                self.wait_and_tick()
                                continue

                            # node is already stopped
                            elif 'SERVER_STOPPED' in str(feedback):
                                print("- skipped - node is already stopped")

                            # fatal error
                            else:
                                print("Error: unable to stop node '{}'!".format(nodeName))
                                print(str(feedback))
                                exit(-1)

                        # quit the loop
                        break

    @staticmethod
    def wait_and_tick(tick=3):
        """Animate the screen while delaying next call to the API"""

        sys.stdout.write('-\r')
        sys.stdout.flush()
        time.sleep(tick)
        sys.stdout.write('\\\r')
        sys.stdout.flush()
        time.sleep(tick)
        sys.stdout.write('|\r')
        sys.stdout.flush()
        time.sleep(tick)
        sys.stdout.write('/\r')
        sys.stdout.flush()
        time.sleep(tick)
        sys.stdout.write(' \r')



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# standard libraries
import os

# yaml for descriptions - http://pyyaml.org/wiki/PyYAMLDocumentation
import yaml

# Apache Libcloud - https://libcloud.readthedocs.org/en/latest
from libcloud.compute.providers import get_driver
from libcloud.compute.types import Provider

# other code related to plumbery
from facility import PlumberyFacility


__all__ = ['PlumberyEngine', 'PlumberyBlueprints']


class PlumberyEngine:
    """Cloud automation at Dimension Data with Apache Libcloud

    Plumbery is a convenient tool for infrastructure managers at cloud times.
    It allows for easy and repeatable deployments of various
    fittings, including compute nodes and related storage. It allows also for
    quick dismandling of the fittings when required.
    The idea here is not to compete with respected solutions such as chef or
    puppet. At its name implies, plumbery is targeting pipes and fittings, the
    very basic utility stuff that sophisticated tools can leverage.

    Args:
        fileName (str): the location of the plan for the fittings

    Example::

        from plumbery.engine import PlumberyEngine
        engine = PlumberyEngine('fittings.yaml')
        engine.build_all_blueprints()

    In this example the overall plan, in textual format, is given to the engine
    in preparation of subsequent processing. The software is not trying to guess
    a name by default, so if you do not provide a name, no configuration file
    is loaded. You can load the plan at any stage, or restart the engine
    with an updated plan, by invoking the member function `parse_layout()`

    Note:
        While plumbery is not making assumptions for your configuration files,
        if your infrastructure is simple enough to fit in one single file then
        you are highly encouraged to name it `fittings.yaml`

    Beyond the plan for your fittings, plumbery is also requiring some specific
    credentials to connect to cloud providers. To preserve the confidentiality
    of such information, it is read from the environment, and not from any
    configuration file. Therefore the need for local setup before running
    plumbery. This is part of the installation process.

    Last but not least, plumbery sets the root password of any new server that
    it creates. For obvious security reasons this is not taken from the fittings
    plan but from the environement as well.

    Under Linux, you may want to edit `~/.bash_profile` like this::

        # credentials to access cloud resources from Dimension Data
        export MCP_USERNAME='foo.bar'
        export MCP_PASSWORD='WhatsUpDoc'

        # password to access nodes remotely
        export SHARED_SECRET='*you really want to put a tricky password here*'

    These simple precautions are aiming to protect your infrastructure as much
    as possible. The general goal is to minimize risks induced by exposure to
    your fittings plan. You may lead transformation towards so-called
    infrastructure as code, and for this you will add version control to your
    fittings plan. As a result, plans will be stored in git or equivalent, and
    shared across some people.

    Attributes:
        driver (libcloud.base.NodeDriver):
            A handle to the underlying Apache Libcloud instance

        facilities (list of PlumberyFacility objects):
            Breakdown of the overall plan over multiple facilities. This global
            attribute results from the parsing of the fittings plan.

        safeMode (boolean):
            If True, which is the default, then no actual change
            will be made against the infrastructure. This global attribute
            is coming from the fittings plan.

        sharedSecret (str):
            The main password used during the creation of a new
            node. This attribute is read from the local environment.

    """

    # the Apache Libcloud driver
    driver = None

    # the various facilities where fittings are put under control
    facilities = []

    # in safe mode no change is made to the fittings
    safeMode = True

    # the password to access remote servers
    sharedSecret = None

    def __init__(self, fileName=None):
        """Ignite the plumbering engine"""

        # get libcloud driver for Managed Cloud Platform (MCP) of Dimension Data
        self.driver = get_driver(Provider.DIMENSIONDATA)

        # get API credentials from environment - with bash, edit ~/.bash_profile to export your credentials in local environment
        self.userName = os.getenv('MCP_USERNAME', "Set environment variable MCP_USERNAME with credentials given to you")
        self.userPassword = os.getenv('MCP_PASSWORD', "Set environment variable MCP_PASSWORD with credentials given to you")

        # get root password from environment - with bash, edit ~/.bash_profile to export SHARED_SECRET in local environment
        self.sharedSecret = os.getenv('SHARED_SECRET')
        if self.sharedSecret is None or len(self.sharedSecret) < 3:
            print("Error: set environment variable SHARED_SECRET with the password to access nodes remotely!")
            exit(-1)

        # load the plan
        if fileName:
            self.parse_layout(fileName)

    def build_all_blueprints(self):
        """Build all blueprints

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to build all blueprints there.

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_all_blueprints()

        """

        print("Building all blueprints")

        for facility in self.facilities:
            facility.focus()
            facility.build_all_blueprints()

    def build_blueprint(self, name):
        """Build a named blueprint

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to build one single blueprint there.

        Args:
            name (str): the name of the blueprint to deploy

        Example::

            from plumbery.engine import PlumberyEngine
            PlumberyEngine('fittings.yaml').build_blueprints('sql')

        """

        print("Building blueprint '{}'".format(name))

        for facility in self.facilities:
            facility.focus()
            facility.build_blueprint(name)

    def destroy_all_nodes(self):
        """Destroy all nodes

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to destroy all nodes there.

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        print("Destroying nodes from all blueprints")

        for facility in self.facilities:
            facility.focus()
            facility.destroy_all_nodes()

    def destroy_nodes(self, name):
        """Destroy nodes

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to destroy nodes from one single blueprint.

        Args:
            name (str): the name of the blueprint to destroy

        Note:
            Running nodes are always preserved from destruction.
            Therefore the need to stop nodes, in a separate command, before
            they can be actually destroyed.

        """

        print("Destroying nodes from blueprint '{}'".format(name))

        for facility in self.facilities:
            facility.focus()
            facility.destroy_nodes(name)

    def parse_layout(self, fileName=None):
        """Read the fittings plan

        The fittings plan is expected to follow YAML specifications, and its
        structure has to follow some rules described here.

        Args:
            fileName (str): the location of the plan for the fittings

        An example of a minimum fittings plan::

            ---
            # Frankfurt in Europe
            locationId: EU6
            regionId: dd-eu

            blueprints:

              - myBluePrint:
                  domain:
                    name: myDC
                  ethernet:
                    name: myVLAN
                    subnet: 10.1.10.0
                  nodes:
                    - myServer:

        In this example, the plan is to deploy a single node in the data centre
        at Frankfurt, in Europe. The node `myServer` will be placed in a
        network named `myVLAN`, and the network will be part of a network
        domain acting as a virtual data centre, `myDC`. The blueprint has a
        name, `myBluePrint`, so that it can be handled independently from
        other blueprints.
        """

        # get file name from the environment, or use default name
        if not fileName:
            fileName = os.getenv('PLUMBERY', 'fittings.yaml')

        # maybe file cannot be read or YAML is broken
        try:
            with open(fileName, 'r') as stream:
                documents = yaml.load_all(stream)

                # first document provides meta information
                document = documents.next()

                if 'safeMode' in document:
                    self.safeMode = document['safeMode']

                # one document per facility
                for document in documents:
                    facility = PlumberyFacility(self, PlumberyBlueprints(**document))
                    self.facilities.append(facility)

        except Exception as feedback:
            print("Error: unable to load file '{}'!".format(fileName))
            print(str(feedback))
            exit(-1)

        # are we in safe mode?
        if self.safeMode:
            print("Running in safe mode - no actual change will be made to the fittings")

    def start_all_nodes(self):
        """Start all nodes

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to start all nodes there.

        This function has no effect on nodes that are already up and running.

        """

        print("Starting nodes from all blueprints")

        for facility in self.facilities:
            facility.focus()
            facility.start_all_nodes()

    def start_nodes(self, name):
        """Start nodes

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to start nodes from one single blueprint.

        Args:
            name (str): the name of the blueprint to start

        This function has no effect on nodes that are already up and running.

        """

        print("Starting nodes from blueprint '{}'".format(name))

        for facility in self.facilities:
            facility.focus()
            facility.start_nodes(name)

    def stop_all_nodes(self):
        """Stop all nodes

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to stop all nodes there.

        This function has no effect on nodes that are already stopped.

        """

        print("Stopping nodes from all blueprints")

        for facility in self.facilities:
            facility.focus()
            facility.stop_all_nodes()

    def stop_nodes(self, name):
        """Stop nodes

        This function will check all facilities, one at a time and in the order
        defined in fittings plan, to stop nodes from one single blueprint.

        Args:
            name (str): the name of the blueprint to stop

        This function has no effect on nodes that are already stopped.

        """

        print("Stopping nodes from blueprint '{}'".format(name))

        for facility in self.facilities:
            facility.focus()
            facility.stop_nodes(name)


class PlumberyBlueprints:
    """Describe fittings plan for one facility"""

    # turn a dictionary to an object
    def __init__(self, **entries):
        self.__dict__.update(entries)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# standard libraries
import os

from libcloud.compute.ssh import SSHClient


__all__ = ['PlumberyPolisher']


class PlumberyPolisher:
    """Make new appliances shine"""

    @classmethod
    def from_shelf(cls, polishId):
        """
        Pick up a polisher from the shelf

        :param polishId: name of the polisher to use, e.g., ``spit``
        :type polishId: ``str``

        :return: instance of a polisher ready to use
        :rtype: :class:`PlumberyPolisher`
        """

        # we may not find a suitable polisher
        try:
            moduleName = 'polishers.' + polishId
            polisherName = polishId.capitalize() + 'Polisher'

            polisherModule = __import__(moduleName, globals(), locals(), [polisherName])
            polisherClass = getattr(polisherModule, polisherName)
            return polisherClass()

        except Exception as feedback:
            print("Error: unable to load polisher '{}'!".format(polishId))
            print(str(feedback))
            return None

    def shine_node(self, node):
        """
        Rub it until it shines

        :param node: the node to be polished
        :type node: :class:`libcloud.compute.base.Node`
        """

        raise NotImplementedError("Error: do not know how to polish '{}'".format(node.name))

    @classmethod
    def rub_node(cls, node, rubs):
        """
        Communicate with the node over SSH
        """

        # get root password from environment SHARED_SECRET
        # with bash, edit ~/.bash_profile to export in local environment
        sharedSecret = os.getenv('SHARED_SECRET', 'whatSUpDoc')

        # use libcloud to communicate to remote nodes
        session = SSHClient(hostname=node.private_ips[0],
                            port=22,
                            username='root',
                            password=sharedSecret,
                            key_files=None,
                            timeout=15)

        # end to end private connectivity is required to succeed
        try:
            session.connect()
            node = rubs.run(node, session)

        except Exception as feedback:
            print("Error: unable to rub '{}' at '{}'!".format(node.name,
                                                             node.private_ips[0]))
            print(str(feedback))
            result = False

        else:
            result = True

        # closing could kill everything as well
        try:
            session.close()
        except:
            pass

        return result



#!/usr/bin/env python

"""
Tests for `plumbery` module.
"""

import os
import unittest

from plumbery.engine import PlumberyEngine

os.environ['SHARED_SECRET'] = "WhatsUpDoc"

class TestPlumberyEngine(unittest.TestCase):

    def setUp(self):
        self.engine = PlumberyEngine()

    def tearDown(self):
        self.engine = None

    def test_000(self):
        pass


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python

"""
Tests for `domain` module.
"""

import os
import unittest

from plumbery.domain import PlumberyDomain


class FakePlumbery:

    safeMode = False


class FakeRegion:

    def ex_create_network_domain(self, location, name, service_plan, description):
        return True

    def ex_create_vlan(self, network_domain, name, private_ipv4_base_address, description):
        return True

    def ex_list_network_domains(self, location):
        return []

    def ex_list_vlans(self, location, network_domain):
        return []


class FakeFacility:

    plumbery = FakePlumbery()

    region = FakeRegion()

    location = 1


fakeBluePrint = {'domain': {'name': 'fake',
                            'service': 'ADVANCED',
                            'description': '#vdc1'},
                'ethernet': {'name': 'fake',
                            'subnet': '10.0.10.0',
                            'description': '#vdc1'}}

class TestPlumberyDomain(unittest.TestCase):

    def setUp(self):
        facility = FakeFacility()
        self.domain = PlumberyDomain(facility=facility)

    def tearDown(self):
        self.domain = None

    def test_000(self):
        self.domain.build(fakeBluePrint)


if __name__ == '__main__':
    import sys
    sys.exit(unittest.main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from os.path import join as pjoin
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


requirements = ['apache-libcloud']

test_requirements = [
    # TODO: put package test requirements here
]

# get version from package itself
def get_version():
    version = None
    sys.path.insert(0, pjoin(os.getcwd()))
    from plumbery import __version__
    version = __version__
    sys.path.pop(0)
    return version

# get description from README.rst
def get_long_description():
    description = ''
    with open('README.rst') as stream:
        description = stream.read()
    return description

setup(
    name='plumbery',
    version=get_version(),
    description="Cloud automation at Dimension Data with Apache Libcloud",
    long_description=get_long_description(),
    author="Bernard Paques",
    author_email='bernard.paques@gmail.com',
    url='https://github.com/bernard357/plumbery',
    packages=['plumbery'],
    package_dir={'plumbery': 'plumbery'},
    include_package_data=True,
    install_requires=requirements,
    license='Apache License (2.0)',
    zip_safe=False,
    keywords='plumbery',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 2.7',
#        'Programming Language :: Python :: 3',
#        'Programming Language :: Python :: 3.0',
#        'Programming Language :: Python :: 3.1',
#        'Programming Language :: Python :: 3.2',
#        'Programming Language :: Python :: 3.3',
#        'Programming Language :: Python :: 3.4',
#        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    test_suite='tests',
    tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Update encrypted deploy password in Travis config file
"""


from __future__ import print_function
import base64
import json
import os
from getpass import getpass
import yaml
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15


try:
    from urllib import urlopen
except:
    from urllib.request import urlopen


GITHUB_REPO = 'bernard357/plumbery'
TRAVIS_CONFIG_FILE = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), '.travis.yml')


def load_key(pubkey):
    """Load public RSA key, with work-around for keys using
    incorrect header/footer format.

    Read more about RSA encryption with cryptography:
    https://cryptography.io/latest/hazmat/primitives/asymmetric/rsa/
    """
    try:
        return load_pem_public_key(pubkey.encode(), default_backend())
    except ValueError:
        # workaround for https://github.com/travis-ci/travis-api/issues/196
        pubkey = pubkey.replace('BEGIN RSA', 'BEGIN').replace('END RSA', 'END')
        return load_pem_public_key(pubkey.encode(), default_backend())


def encrypt(pubkey, password):
    """Encrypt password using given RSA public key and encode it with base64.

    The encrypted password can only be decrypted by someone with the
    private key (in this case, only Travis).
    """
    key = load_key(pubkey)
    encrypted_password = key.encrypt(password, PKCS1v15())
    return base64.b64encode(encrypted_password)


def fetch_public_key(repo):
    """Download RSA public key Travis will use for this repo.

    Travis API docs: http://docs.travis-ci.com/api/#repository-keys
    """
    keyurl = 'https://api.travis-ci.org/repos/{0}/key'.format(repo)
    data = json.loads(urlopen(keyurl).read())
    if 'key' not in data:
        errmsg = "Could not find public key for repo: {}.\n".format(repo)
        errmsg += "Have you already added your GitHub repo to Travis?"
        raise ValueError(errmsg)
    return data['key']


def prepend_line(filepath, line):
    """Rewrite a file adding a line to its beginning.
    """
    with open(filepath) as f:
        lines = f.readlines()

    lines.insert(0, line)

    with open(filepath, 'w') as f:
        f.writelines(lines)


def load_yaml_config(filepath):
    with open(filepath) as f:
        return yaml.load(f)


def save_yaml_config(filepath, config):
    with open(filepath, 'w') as f:
        yaml.dump(config, f, default_flow_style=False)


def update_travis_deploy_password(encrypted_password):
    """Update the deploy section of the .travis.yml file
    to use the given encrypted password.
    """
    config = load_yaml_config(TRAVIS_CONFIG_FILE)

    config['deploy']['password'] = dict(secure=encrypted_password)

    save_yaml_config(TRAVIS_CONFIG_FILE, config)

    line = ('# This file was autogenerated and will overwrite'
            ' each time you run travis_pypi_setup.py\n')
    prepend_line(TRAVIS_CONFIG_FILE, line)


def main(args):
    public_key = fetch_public_key(args.repo)
    password = args.password or getpass('PyPI password: ')
    update_travis_deploy_password(encrypt(public_key, password))
    print("Wrote encrypted password to .travis.yml -- you're ready to deploy")


if '__main__' == __name__:
    import argparse
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--repo', default=GITHUB_REPO,
                        help='GitHub repo (default: %s)' % GITHUB_REPO)
    parser.add_argument('--password',
                        help='PyPI password (will prompt if not provided)')

    args = parser.parse_args()
    main(args)



# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


__all__ = ['PlumberyDomain']


class PlumberyDomain:
    """Cloud automation for a network domain

    A network domain is similar to a virtual data center. It is a secured
    container for multiple nodes.

    Args:
        facility (PlumberyFacility): the underlying physical facility

    Example::

        from plumbery.domain import PlumberyDomain
        domain = PlumberyDomain(facility=facility)
        domain.build_blueprint(blueprint=blueprint)

    In this example a domain is initialised at the given facility, and then
    it is asked to create the pipes and the plumbery mentioned in the
    provided blueprint. This is covering solely the network and the security,
    not the nodes themselves.

    Attributes:
        facility (PlumberyFacility): a handle to the physical facility where
            network domains are implemented

    """

    # the physical data center
    facility = None

    def __init__(self, facility=None):
        """Put network domains in context"""

        # handle to parent parameters and functions
        self.facility = facility
        self.region = facility.region
        self.plumbery = facility.plumbery

    def build(self, blueprint):
        """Create network domain if it does not exist

        Args:
            blueprint (dict): the various attributes of the target fittings

        """

        # check the target network domain
        if 'domain' not in blueprint or type(blueprint['domain']) is not dict:
            print("Error: no network domain has been defined for the blueprint '{}'!".format(blueprint['target']))
            exit(-1)

        # seek for an existing network domain with this name
        domainName = blueprint['domain']['name']
        self.domain = None
        for self.domain in self.region.ex_list_network_domains(location=self.facility.location):
            if self.domain.name == domainName:
                print("Network domain '{}' already exists".format(domainName))
                break

        # create a network domain if needed
        if self.domain is None or self.domain.name != domainName:

            if self.plumbery.safeMode:
                print("Would have created network domain '{}' if not in safe mode".format(domainName))
                exit(0)

            else:
                print("Creating network domain '{}'".format(domainName))

                # the description attribute is a smart way to tag resources
                description = '#plumbery'
                if 'description' in blueprint['domain']:
                    description = blueprint['domain']['description'] + ' #plumbery'

                # level of service
                service = 'ESSENTIALS'
                if 'service' in blueprint['domain']:
                    service = blueprint['domain']['service']

                # we may have to wait for busy resources
                while True:

                    try:
                        self.domain = self.region.ex_create_network_domain(
                            location=self.facility.location,
                            name=domainName,
                            service_plan=service,
                            description=description)
                        print("- in progress")

                    except Exception as feedback:

                        # resource is busy, wait a bit and retry
                        if 'RESOURCE_BUSY' in str(feedback):
                            self.facility.wait_and_tick()
                            continue

                        # fatal error
                        else:
                            print("Error: unable to create network domain '{}'!".format(domainName))
                            print(str(feedback))
                            exit(-1)

                    # quit the loop
                    break

        # check name of the target network
        if 'ethernet' not in blueprint or type(blueprint['ethernet']) is not dict:
            print("Error: no ethernet network has been defined for the blueprint '{}'!".format(blueprint['target']))
            exit(-1)

        # check addresses to use for the target network
        if 'subnet' not in blueprint['ethernet']:
            print("Error: no IPv4 subnet (e.g., '10.0.34.0') as been defined for the blueprint '{}'!".format(blueprint['target']))
            exit(-1)

        # seek for an existing network with this name
        networkName = blueprint['ethernet']['name']
        self.network = None
        for self.network in self.region.ex_list_vlans(location=self.facility.location, network_domain=self.domain):
            if self.network.name == networkName:
                print("Ethernet network '{}' already exists".format(networkName))
                break

        # create a network if needed
        if self.network is None or self.network.name != networkName:

            if self.plumbery.safeMode:
                print("Would have created Ethernet network '{}' if not in safe mode".format(networkName))
                exit(0)

            else:
                print("Creating Ethernet network '{}'".format(networkName))

                # the description attribute is a smart way to tag resources
                description = '#plumbery'
                if 'description' in blueprint['ethernet']:
                    description = blueprint['ethernet']['description'] + ' #plumbery'

                # we may have to wait for busy resources
                while True:

                    try:
                        self.network = self.region.ex_create_vlan(
                            network_domain=self.domain,
                            name=networkName,
                            private_ipv4_base_address=blueprint['ethernet']['subnet'],
                            description=description)
                        print("- in progress")

                    except Exception as feedback:

                        # resource is busy, wait a bit and retry
                        if 'RESOURCE_BUSY' in str(feedback):
                            self.facility.wait_and_tick()
                            continue

                        # fatal error
                        else:
                            print("Error: unable to create Ethernet network '{}'!".format(networkName))
                            print(str(feedback))
                            exit(-1)

                    # quit the loop
                    break
